
#ifndef __WFX_MATH_MATRIX4_H__
#define __WFX_MATH_MATRIX4_H__

#include <WFX/WFXIncludes.h>
#include <WFX/math/Vector3.h>
#include <WFX/math/Matrix3.h>

namespace WFX {

    namespace math {

        /**
         * a 4x4 Matrix
         */
        class Matrix4 {
            public:
                /**
                 * ctor
                 */
                inline Matrix4(
                        Real m00, Real m01, Real m02, Real m03,
                        Real m10, Real m11, Real m12, Real m13,
                        Real m20, Real m21, Real m22, Real m23,
                        Real m30, Real m31, Real m32, Real m33 )
                {
                    m[0][0] = m00;
                    m[0][1] = m01;
                    m[0][2] = m02;
                    m[0][3] = m03;
                    m[1][0] = m10;
                    m[1][1] = m11;
                    m[1][2] = m12;
                    m[1][3] = m13;
                    m[2][0] = m20;
                    m[2][1] = m21;
                    m[2][2] = m22;
                    m[2][3] = m23;
                    m[3][0] = m30;
                    m[3][1] = m31;
                    m[3][2] = m32;
                    m[3][3] = m33;
                }

                /** 
                 * contructs a Matrix4 from a rotation Matrix3, a translation
                 * Vector3 and a scaling factor
                 */
                Matrix4(const Matrix3& rot, const Vector3& trans, Real scale);

                Matrix4() {
                    memset(m,0,16*sizeof(Real));
                }
                inline Real* operator [] ( unsigned iRow )
                {
                    assert( iRow < 4 );
                    return m[iRow];
                }

                inline const Real *const operator [] ( unsigned iRow ) const
                {
                    assert( iRow < 4 );
                    return m[iRow];
                }

                inline Matrix4 concatenate(const Matrix4 &m2) const {
                    Matrix4 r;

                    r.m[0][0] = m[0][0] * m2.m[0][0] + m[0][1] * m2.m[1][0] + m[0][2] * m2.m[2][0] + m[0][3] * m2.m[3][0];
                    r.m[0][1] = m[0][0] * m2.m[0][1] + m[0][1] * m2.m[1][1] + m[0][2] * m2.m[2][1] + m[0][3] * m2.m[3][1];
                    r.m[0][2] = m[0][0] * m2.m[0][2] + m[0][1] * m2.m[1][2] + m[0][2] * m2.m[2][2] + m[0][3] * m2.m[3][2];
                    r.m[0][3] = m[0][0] * m2.m[0][3] + m[0][1] * m2.m[1][3] + m[0][2] * m2.m[2][3] + m[0][3] * m2.m[3][3];

                    r.m[1][0] = m[1][0] * m2.m[0][0] + m[1][1] * m2.m[1][0] + m[1][2] * m2.m[2][0] + m[1][3] * m2.m[3][0];
                    r.m[1][1] = m[1][0] * m2.m[0][1] + m[1][1] * m2.m[1][1] + m[1][2] * m2.m[2][1] + m[1][3] * m2.m[3][1];
                    r.m[1][2] = m[1][0] * m2.m[0][2] + m[1][1] * m2.m[1][2] + m[1][2] * m2.m[2][2] + m[1][3] * m2.m[3][2];
                    r.m[1][3] = m[1][0] * m2.m[0][3] + m[1][1] * m2.m[1][3] + m[1][2] * m2.m[2][3] + m[1][3] * m2.m[3][3];

                    r.m[2][0] = m[2][0] * m2.m[0][0] + m[2][1] * m2.m[1][0] + m[2][2] * m2.m[2][0] + m[2][3] * m2.m[3][0];
                    r.m[2][1] = m[2][0] * m2.m[0][1] + m[2][1] * m2.m[1][1] + m[2][2] * m2.m[2][1] + m[2][3] * m2.m[3][1];
                    r.m[2][2] = m[2][0] * m2.m[0][2] + m[2][1] * m2.m[1][2] + m[2][2] * m2.m[2][2] + m[2][3] * m2.m[3][2];
                    r.m[2][3] = m[2][0] * m2.m[0][3] + m[2][1] * m2.m[1][3] + m[2][2] * m2.m[2][3] + m[2][3] * m2.m[3][3];

                    r.m[3][0] = m[3][0] * m2.m[0][0] + m[3][1] * m2.m[1][0] + m[3][2] * m2.m[2][0] + m[3][3] * m2.m[3][0];
                    r.m[3][1] = m[3][0] * m2.m[0][1] + m[3][1] * m2.m[1][1] + m[3][2] * m2.m[2][1] + m[3][3] * m2.m[3][1];
                    r.m[3][2] = m[3][0] * m2.m[0][2] + m[3][1] * m2.m[1][2] + m[3][2] * m2.m[2][2] + m[3][3] * m2.m[3][2];
                    r.m[3][3] = m[3][0] * m2.m[0][3] + m[3][1] * m2.m[1][3] + m[3][2] * m2.m[2][3] + m[3][3] * m2.m[3][3];

                    return r;
                }

                inline Matrix4 operator * ( const Matrix4 &m2 ) const {
                    return concatenate( m2 );
                }

                /** Vector transformation using '*'.
                 * @remarks
                 * Transforms the given 3-D vector by the matrix, projecting the 
                 * result back into <i>w</i> = 1.
                 * @note
                 * This means that the initial <i>w</i> is considered to be 1.0,
                 * and then all the tree elements of the resulting 3-D vector are
                 * divided by the resulting <i>w</i>.
                 */
                inline Vector3 operator * ( const Vector3 &v ) const {
                    Vector3 r;

                    Real fInvW = 1.0 / ( m[3][0] * v.X + m[3][1] * v.Y + m[3][2] * v.Z + m[3][3] );

                    r.X = ( m[0][0] * v.X + m[0][1] * v.Y + m[0][2] * v.Z + m[0][3] ) * fInvW;
                    r.Y = ( m[1][0] * v.X + m[1][1] * v.Y + m[1][2] * v.Z + m[1][3] ) * fInvW;
                    r.Z = ( m[2][0] * v.X + m[2][1] * v.Y + m[2][2] * v.Z + m[2][3] ) * fInvW;

                    return r;
                }

                /** Matrix addition.
                 */
                inline Matrix4 operator + ( const Matrix4 &m2 ) const {
                    Matrix4 r;

                    r.m[0][0] = m[0][0] + m2.m[0][0];
                    r.m[0][1] = m[0][1] + m2.m[0][1];
                    r.m[0][2] = m[0][2] + m2.m[0][2];
                    r.m[0][3] = m[0][3] + m2.m[0][3];

                    r.m[1][0] = m[1][0] + m2.m[1][0];
                    r.m[1][1] = m[1][1] + m2.m[1][1];
                    r.m[1][2] = m[1][2] + m2.m[1][2];
                    r.m[1][3] = m[1][3] + m2.m[1][3];

                    r.m[2][0] = m[2][0] + m2.m[2][0];
                    r.m[2][1] = m[2][1] + m2.m[2][1];
                    r.m[2][2] = m[2][2] + m2.m[2][2];
                    r.m[2][3] = m[2][3] + m2.m[2][3];

                    r.m[3][0] = m[3][0] + m2.m[3][0];
                    r.m[3][1] = m[3][1] + m2.m[3][1];
                    r.m[3][2] = m[3][2] + m2.m[3][2];
                    r.m[3][3] = m[3][3] + m2.m[3][3];

                    return r;
                }

                /** Matrix subtraction.
                 *         */
                inline Matrix4 operator - ( const Matrix4 &m2 ) const {
                    Matrix4 r;
                    r.m[0][0] = m[0][0] - m2.m[0][0];
                    r.m[0][1] = m[0][1] - m2.m[0][1];
                    r.m[0][2] = m[0][2] - m2.m[0][2];
                    r.m[0][3] = m[0][3] - m2.m[0][3];

                    r.m[1][0] = m[1][0] - m2.m[1][0];
                    r.m[1][1] = m[1][1] - m2.m[1][1];
                    r.m[1][2] = m[1][2] - m2.m[1][2];
                    r.m[1][3] = m[1][3] - m2.m[1][3];

                    r.m[2][0] = m[2][0] - m2.m[2][0];
                    r.m[2][1] = m[2][1] - m2.m[2][1];
                    r.m[2][2] = m[2][2] - m2.m[2][2];
                    r.m[2][3] = m[2][3] - m2.m[2][3];

                    r.m[3][0] = m[3][0] - m2.m[3][0];
                    r.m[3][1] = m[3][1] - m2.m[3][1];
                    r.m[3][2] = m[3][2] - m2.m[3][2];
                    r.m[3][3] = m[3][3] - m2.m[3][3];

                    return r;
                }

                /** Tests 2 matrices for equality.
                 *         */
                inline bool operator == ( const Matrix4& m2 ) const {
                    if (
                            m[0][0] != m2.m[0][0] || m[0][1] != m2.m[0][1] || m[0][2] != m2.m[0][2] || m[0][3] != m2.m[0][3] ||
                            m[1][0] != m2.m[1][0] || m[1][1] != m2.m[1][1] || m[1][2] != m2.m[1][2] || m[1][3] != m2.m[1][3] ||
                            m[2][0] != m2.m[2][0] || m[2][1] != m2.m[2][1] || m[2][2] != m2.m[2][2] || m[2][3] != m2.m[2][3] ||
                            m[3][0] != m2.m[3][0] || m[3][1] != m2.m[3][1] || m[3][2] != m2.m[3][2] || m[3][3] != m2.m[3][3] 
                        ) return false;
                    return true;
                }

                /** Tests 2 matrices for inequality.
                 *         */
                inline bool operator != ( Matrix4& m2 ) const {
                    if (
                            m[0][0] != m2.m[0][0] || m[0][1] != m2.m[0][1] || m[0][2] != m2.m[0][2] || m[0][3] != m2.m[0][3] ||
                            m[1][0] != m2.m[1][0] || m[1][1] != m2.m[1][1] || m[1][2] != m2.m[1][2] || m[1][3] != m2.m[1][3] ||
                            m[2][0] != m2.m[2][0] || m[2][1] != m2.m[2][1] || m[2][2] != m2.m[2][2] || m[2][3] != m2.m[2][3] ||
                            m[3][0] != m2.m[3][0] || m[3][1] != m2.m[3][1] || m[3][2] != m2.m[3][2] || m[3][3] != m2.m[3][3] 
                        ) return true;
                    return false;
                }

                inline Matrix4 operator*(Real scalar) {
                    return Matrix4(
                            scalar*m[0][0], scalar*m[0][1], scalar*m[0][2], scalar*m[0][3],
                            scalar*m[1][0], scalar*m[1][1], scalar*m[1][2], scalar*m[1][3],
                            scalar*m[2][0], scalar*m[2][1], scalar*m[2][2], scalar*m[2][3],
                            scalar*m[3][0], scalar*m[3][1], scalar*m[3][2], scalar*m[3][3]);
                }

                
                /** Assignment from 3x3 matrix.
                 */
                inline void operator =(const Matrix3& mat3) {
                    m[0][0] = mat3.m[0][0]; m[0][1] = mat3.m[0][1]; m[0][2] = mat3.m[0][2];
                    m[1][0] = mat3.m[1][0]; m[1][1] = mat3.m[1][1]; m[1][2] = mat3.m[1][2];
                    m[2][0] = mat3.m[2][0]; m[2][1] = mat3.m[2][1]; m[2][2] = mat3.m[2][2];
                }

                inline Matrix4 transpose(void) const {
                    return Matrix4(m[0][0], m[1][0], m[2][0], m[3][0],
                            m[0][1], m[1][1], m[2][1], m[3][1],
                            m[0][2], m[1][2], m[2][2], m[3][2],
                            m[0][3], m[1][3], m[2][3], m[3][3]);
                }

              
                static const Matrix4 ZERO;
                static const Matrix4 IDENTITY;
                
            protected:
                /** elements */
                Real m[4][4];
                
        };

    }

}


#endif //include guard


